from typing import Optional

import numpy as np
import scipy.special
import torch
from scipy.optimize import *
from torch import Tensor, nn

from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.fn.dirac_delta import dirac_delta
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSOR_OPERABLE
from analogvnn.utils.to_tensor_parameter import to_float_tensor, to_nongrad_parameter

__all__ = ['PoissonNoise']


class PoissonNoise(Layer, BackwardIdentity):
    """Implements the Poisson noise function.

    Attributes:
        scale (nn.Parameter): the scale of the Poisson noise function.
        max_leakage (nn.Parameter): the maximum leakage of the Poisson noise.
        precision (nn.Parameter): the precision of the Poisson noise.
    """
    __constants__ = ['scale', 'max_leakage', 'precision']
    scale: nn.Parameter
    max_leakage: nn.Parameter
    precision: nn.Parameter

    def __init__(
            self,
            scale: Optional[float] = None,
            max_leakage: Optional[float] = None,
            precision: Optional[int] = None,
    ):
        """Initializes the Poisson noise function.

        Args:
            scale (Optional[float]): the scale of the Poisson noise function.
            max_leakage (Optional[float]): the maximum leakage of the Poisson noise.
            precision (Optional[int]): the precision of the Poisson noise.
        """
        super(PoissonNoise, self).__init__()

        if (scale is None) + (max_leakage is None) + (precision is None) != 1:
            raise ValueError("only 2 out of 3 arguments are needed (scale, max_leakage, precision)")

        scale, max_leakage, precision = to_float_tensor(scale, max_leakage, precision)

        if scale is None and max_leakage is not None and precision is not None:
            scale = self.calc_scale(max_leakage, precision)

        if precision is None and scale is not None and max_leakage is not None:
            precision = self.calc_precision(scale, max_leakage)

        if max_leakage is None and scale is not None and precision is not None:
            max_leakage = self.calc_max_leakage(scale, precision)

        self.scale, self.max_leakage, self.precision = to_nongrad_parameter(scale, max_leakage, precision)

        if self.rate_factor < 1.:
            raise ValueError("rate_factor must be >= 1 (scale * precision >= 1)")

    @staticmethod
    def calc_scale(max_leakage: TENSOR_OPERABLE, precision: TENSOR_OPERABLE, max_check=10000) -> TENSOR_OPERABLE:
        """Calculates the scale using the maximum leakage and the precision.

        Args:
            max_leakage (TENSOR_OPERABLE): the maximum leakage of the Poisson noise.
            precision (TENSOR_OPERABLE): the precision of the Poisson noise.
            max_check (int): the maximum value to check for the scale.

        Returns:
            TENSOR_OPERABLE: the scale of the Poisson noise function.
        """
        max_leakage = float(max_leakage)
        precision = float(precision)
        r, _ = toms748(
            lambda s: PoissonNoise.calc_max_leakage(s, precision) - max_leakage,
            a=0,
            b=max_check,
            maxiter=10000
        )
        return r

    @staticmethod
    def calc_precision(scale: TENSOR_OPERABLE, max_leakage: TENSOR_OPERABLE, max_check=2 ** 16) -> TENSOR_OPERABLE:
        """Calculates the precision using the scale and the maximum leakage.

        Args:
            scale (TENSOR_OPERABLE): the scale of the Poisson noise function.
            max_leakage (TENSOR_OPERABLE): the maximum leakage of the Poisson noise.
            max_check (int): the maximum value to check for the precision.

        Returns:
            TENSOR_OPERABLE: the precision of the Poisson noise.
        """
        max_leakage = float(max_leakage)
        scale = float(scale)
        r, _ = toms748(
            lambda p: PoissonNoise.calc_max_leakage(scale, p) - max_leakage,
            a=1,
            b=max_check,
            maxiter=10000
        )
        return r

    @staticmethod
    def calc_max_leakage(scale: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """Calculates the maximum leakage using the scale and the precision.

        Args:
            scale (TENSOR_OPERABLE): the scale of the Poisson noise function.
            precision (TENSOR_OPERABLE): the precision of the Poisson noise.

        Returns:
            TENSOR_OPERABLE: the maximum leakage of the Poisson noise.
        """
        return 1 - (
                PoissonNoise.static_cdf(x=1. + 1 / (2 * precision), rate=1., scale_factor=scale * precision)
                - PoissonNoise.static_cdf(x=1. - 1 / (2 * precision), rate=1., scale_factor=scale * precision)
        )

    @staticmethod
    def static_cdf(x: TENSOR_OPERABLE, rate: TENSOR_OPERABLE, scale_factor: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """Calculates the cumulative distribution function of the Poisson noise.

        Args:
            x (TENSOR_OPERABLE): the input of the Poisson noise.
            rate (TENSOR_OPERABLE): the rate of the Poisson noise.
            scale_factor (TENSOR_OPERABLE): the scale factor of the Poisson noise.

        Returns:
            TENSOR_OPERABLE: the cumulative distribution function of the Poisson noise.
        """
        if np.isclose(rate, np.zeros_like(rate)):
            return np.ones_like(x)

        x = np.abs(x) * scale_factor
        rate = np.abs(rate) * scale_factor
        return scipy.special.gammaincc(x + 1, rate)

    @staticmethod
    def staticmethod_leakage(scale: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """Calculates the leakage of the Poisson noise using the scale and the precision.

        Args:
            scale (TENSOR_OPERABLE): the scale of the Poisson noise function.
            precision (TENSOR_OPERABLE): the precision of the Poisson noise.

        Returns:
            TENSOR_OPERABLE: the leakage of the Poisson noise.
        """
        cdf_domain = np.linspace(-1, 1, int(2 * precision + 1), dtype=float)
        correctness = 0

        for i in cdf_domain:
            min_i = i - 1 / (2 * precision)
            max_i = i + 1 / (2 * precision)

            if np.isclose(i, 1.):
                max_i = 1.
            if np.isclose(i, -1.):
                min_i = -1.

            if np.isclose(i, 0.):
                correctness += 1
            else:
                correctness += PoissonNoise.static_cdf(
                    max(abs(max_i), abs(min_i)), rate=i, scale_factor=scale * precision
                )
                correctness -= PoissonNoise.static_cdf(
                    min(abs(max_i), abs(min_i)), rate=i, scale_factor=scale * precision
                )

        correctness /= cdf_domain.size - 1
        return 1 - correctness

    @property
    def leakage(self) -> float:
        """The leakage of the Poisson noise.

        Returns:
            float: the leakage of the Poisson noise.
        """
        return self.staticmethod_leakage(scale=float(self.scale), precision=int(self.precision))

    @property
    def rate_factor(self) -> Tensor:
        """The rate factor of the Poisson noise.

        Returns:
            Tensor: the rate factor of the Poisson noise.
        """
        return self.precision * self.scale

    def pdf(self, x: Tensor, rate: Tensor) -> Tensor:
        """Calculates the probability density function of the Poisson noise.

        Args:
            x (Tensor): the input of the Poisson noise.
            rate (Tensor): the rate of the Poisson noise.

        Returns:
            Tensor: the probability density function of the Poisson noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        rate = rate if isinstance(rate, Tensor) else torch.tensor(rate, requires_grad=False)

        if torch.isclose(rate, torch.zeros_like(rate)):
            return dirac_delta(x)

        return torch.exp(self.log_prob(x=x, rate=rate))

    def log_prob(self, x: Tensor, rate: Tensor) -> Tensor:
        """Calculates the log probability of the Poisson noise.

        Args:
            x (Tensor): the input of the Poisson noise.
            rate (Tensor): the rate of the Poisson noise.

        Returns:
            Tensor: the log probability of the Poisson noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        rate = rate if isinstance(rate, Tensor) else torch.tensor(rate, requires_grad=False)

        x = torch.abs(x) * self.rate_factor
        rate = torch.abs(rate) * self.rate_factor
        return x.xlogy(rate) - rate - torch.lgamma(x + 1)

    def cdf(self, x: Tensor, rate: Tensor) -> Tensor:
        """Calculates the cumulative distribution function of the Poisson noise.

        Args:
            x (Tensor): the input of the Poisson noise.
            rate (Tensor): the rate of the Poisson noise.

        Returns:
            Tensor: the cumulative distribution function of the Poisson noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        rate = rate if isinstance(rate, Tensor) else torch.tensor(rate, requires_grad=False)
        return self.static_cdf(x, rate, self.rate_factor)

    def forward(self, x: Tensor) -> Tensor:
        """Adds the Poisson noise to the input.

        Args:
            x (Tensor): the input of the Poisson noise.

        Returns:
            Tensor: the output of the Poisson noise.
        """
        return torch.sign(x) * torch.poisson(torch.abs(x * self.rate_factor)) / self.rate_factor

    def extra_repr(self) -> str:
        """Returns the extra representation of the Poisson noise.

        Returns:
            str: the extra representation of the Poisson noise.
        """
        return f'scale={float(self.scale):.4f}' \
               f', max_leakage={float(self.max_leakage):.4f}' \
               f', leakage={float(self.leakage):.4f}' \
               f', precision={int(self.precision)}'



from typing import Tuple

import torch
from torch import nn
from torch.utils.data import DataLoader

__all__ = ['test']


def test(model: nn.Module, test_loader: DataLoader, test_run: bool = False) -> Tuple[float, float]:
    """Test the model on the test set.

    Args:
        model (torch.nn.Module): the model to test.
        test_loader (DataLoader): the test set.
        test_run (bool): is it a test run.

    Returns:
        tuple: the loss and accuracy of the model on the test set.
    """

    model.eval()
    total_loss = 0
    total_accuracy = 0
    total_size = 0

    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(model.device), target.to(model.device)

            output = model(data)
            loss, accuracy = model.loss(output, target)

            total_loss += loss.item() * len(data)
            total_accuracy += accuracy * len(data)
            total_size += len(data)

            if test_run:
                break

    total_loss /= total_size
    total_accuracy /= total_size
    return total_loss, total_accuracy



from __future__ import annotations

import typing
from typing import Callable, Optional, Tuple

import torch
from torch import optim, Tensor
from torch.utils.data import DataLoader

from analogvnn.fn.test import test
from analogvnn.fn.train import train
from analogvnn.graph.BackwardGraph import BackwardGraph
from analogvnn.graph.ForwardGraph import ForwardGraph
from analogvnn.graph.ModelGraph import ModelGraph
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSORS
from analogvnn.utils.is_cpu_cuda import is_cpu_cuda

if typing.TYPE_CHECKING:
    from analogvnn.utils.TensorboardModelLog import TensorboardModelLog

__all__ = ['Model']


class Model(Layer):
    """Base class for analog neural network models.

    Attributes:
        _compiled (bool): True if the model is compiled.
        tensorboard (TensorboardModelLog): The tensorboard logger of the model.
        graphs (ModelGraph): The graph of the model.
        forward_graph (ForwardGraph): The forward graph of the model.
        backward_graph (BackwardGraph): The backward graph of the model.
        optimizer (optim.Optimizer): The optimizer of the model.
        loss_function (Callable): The loss function of the model.
        accuracy_function (Callable): The accuracy function of the model.
        device (torch.device): The device of the model.
    """
    __constants__ = ['device']

    _compiled: bool

    tensorboard: Optional[TensorboardModelLog]

    graphs: ModelGraph
    forward_graph: ForwardGraph
    backward_graph: BackwardGraph

    optimizer: Optional[optim.Optimizer]
    loss_function: Optional[Callable]
    accuracy_function: Optional[Callable]
    device: torch.device

    def __init__(self, tensorboard_log_dir=None, device=is_cpu_cuda.device):
        """Create a new model.

        Args:
            tensorboard_log_dir (str): The log directory of the tensorboard logger.
            device (torch.device): The device to run the model on.
        """
        super(Model, self).__init__()

        self._compiled = False

        self.tensorboard = None
        if tensorboard_log_dir is not None:
            self.create_tensorboard(tensorboard_log_dir)

        self.graphs = ModelGraph()
        self.forward_graph = self.graphs.forward_graph
        self.backward_graph = self.graphs.backward_graph

        self.optimizer = None
        self.loss_function = None
        self.accuracy_function = None
        self.device = device

    @property
    def use_autograd_graph(self):
        """ Is the autograd graph used for the model.
        Returns:
            bool: If True, the autograd graph is used to calculate the gradients.
        """
        return self.graphs.use_autograd_graph

    @use_autograd_graph.setter
    def use_autograd_graph(self, use_autograd_graph: bool):
        """ Set if the autograd graph is used for the model.

        Args:
            use_autograd_graph (bool): If True, the autograd graph is used to calculate the gradients.
        """
        self.graphs.use_autograd_graph = use_autograd_graph

    def compile(self, device: Optional[torch.device] = None, layer_data: bool = True):
        """Compile the model.

        Args:
            device (torch.device): The device to run the model on.
            layer_data (bool): If True, the layer data is logged.

        Returns:
            Model: The compiled model.
        """
        if device is not None:
            self.device = device

        self.graphs.compile()
        for i in self.modules():
            if isinstance(i, Layer) and i != self:
                i.graphs = self.graphs

        self.to(device=self.device)

        self._compiled = True
        if self.tensorboard is not None:
            self.tensorboard.on_compile(layer_data=layer_data)
        return self

    def forward(self, *inputs: Tensor) -> TENSORS:
        """Forward pass of the model.

        Args:
            *inputs (Tensor): The inputs of the model.

        Returns:
            TENSORS: The output of the model.
        """
        return self.graphs.forward_graph(inputs, self.training)

    @torch.no_grad()
    def backward(self, *inputs: Tensor) -> TENSORS:
        """Backward pass of the model.

        Args:
            *inputs (Tensor): The inputs of the model.

        Returns:
            TENSORS: The output of the model.
        """
        return self.graphs.backward_graph(inputs)

    def loss(self, output: Tensor, target: Tensor) -> Tuple[Tensor, Tensor]:
        """Calculate the loss of the model.

        Args:
            output (Tensor): The output of the model.
            target (Tensor): The target of the model.

        Returns:
            Tuple[Tensor, Tensor]: The loss and the accuracy of the model.

        Raises:
            ValueError: if loss_function is None.
        """
        if self.loss_function is None:
            raise ValueError('loss_function is None')

        loss_result = self.loss_function(output, target)
        if self.training:
            self.graphs.set_loss(loss_result)

        accuracy_result = None
        if self.accuracy_function is not None:
            accuracy_result = self.accuracy_function(output, target)

        return loss_result, accuracy_result

    def train_on(self, train_loader: DataLoader, epoch: int = None, *args, **kwargs) -> Tuple[float, float]:
        """Train the model on the train_loader.

        Args:
            train_loader (DataLoader): The train loader of the model.
            epoch (int): The epoch of the model.
            *args: The arguments of the train function.
            **kwargs: The keyword arguments of the train function.

        Returns:
            Tuple[float, float]: The loss and the accuracy of the model.

        Raises:
            RuntimeError: if model is not compiled.
        """
        if self._compiled is False:
            raise RuntimeError('Model is not compiled')

        train_loss, train_accuracy = train(self, train_loader, epoch, *args, **kwargs)

        if self.tensorboard is not None:
            self.tensorboard.add_graph(train_loader)
            self.tensorboard.register_training(epoch, train_loss, train_accuracy)

        return train_loss, train_accuracy

    def test_on(self, test_loader: DataLoader, epoch: int = None, *args, **kwargs) -> Tuple[float, float]:
        """Test the model on the test_loader.

        Args:
            test_loader (DataLoader): The test loader of the model.
            epoch (int): The epoch of the model.
            *args: The arguments of the test function.
            **kwargs: The keyword arguments of the test function.

        Returns:
            Tuple[float, float]: The loss and the accuracy of the model.

        Raises:
            RuntimeError: if model is not compiled.
        """
        if self._compiled is False:
            raise RuntimeError('Model is not compiled')

        test_loss, test_accuracy = test(self, test_loader, *args, **kwargs)

        if self.tensorboard is not None:
            self.tensorboard.add_graph(test_loader)
            self.tensorboard.register_testing(epoch, test_loss, test_accuracy)

        return test_loss, test_accuracy

    def fit(
            self,
            train_loader: DataLoader,
            test_loader: DataLoader,
            epoch: int = None
    ) -> Tuple[float, float, float, float]:
        """Fit the model on the train_loader and test the model on the test_loader.

        Args:
            train_loader (DataLoader): The train loader of the model.
            test_loader (DataLoader): The test loader of the model.
            epoch (int): The epoch of the model.

        Returns:
            Tuple[float, float, float, float]: The train loss, the train accuracy, the test loss
            and the test accuracy of the model.
        """
        train_loss, train_accuracy = self.train_on(train_loader=train_loader, epoch=epoch)
        test_loss, test_accuracy = self.test_on(test_loader=test_loader, epoch=epoch)
        return train_loss, train_accuracy, test_loss, test_accuracy

    def create_tensorboard(self, log_dir: str) -> TensorboardModelLog:
        """Create a tensorboard.

        Args:
            log_dir (str): The log directory of the tensorboard.

        Raises:
            ImportError: if tensorboard (https://www.tensorflow.org/) is not installed.
        """
        try:
            from analogvnn.utils.TensorboardModelLog import TensorboardModelLog
        except ImportError as e:
            raise ImportError("requires tensorboard https://www.tensorflow.org/") from e

        self.tensorboard = TensorboardModelLog(self, log_dir=log_dir)
        self.subscribe_tensorboard(self.tensorboard)
        return self.tensorboard

    def subscribe_tensorboard(self, tensorboard: TensorboardModelLog):
        """Subscribe the model to the tensorboard.

        Args:
            tensorboard (TensorboardModelLog): The tensorboard of the model.

        Returns:
            Model: self.
        """
        self.tensorboard = tensorboard
        if self._compiled is True:
            self.tensorboard.on_compile()

        return self



from typing import Callable, Union, Sequence

from torch import Tensor

__all__ = ['TENSOR_OPERABLE', 'TENSOR_CALLABLE', 'TENSORS']

TENSOR_OPERABLE = Union[Sequence[Tensor], Tensor, int, float, bool]
""" `TENSOR_OPERABLE` is a type alias for types that can be operated on by a tensor. """

TENSOR_CALLABLE = Callable[[TENSOR_OPERABLE], TENSOR_OPERABLE]
""" `TENSOR_CALLABLE` is a type alias for a function that takes a `TENSOR_OPERABLE` and returns a `TENSOR_OPERABLE`. """

TENSORS = Union[None, Tensor, Sequence[Tensor]]
""" `TENSORS` is a type alias for a tensor or a sequence of tensors. """



import math
from typing import Optional

import numpy as np
import torch
from torch import Tensor, nn

from analogvnn.backward.BackwardIdentity import BackwardIdentity
from analogvnn.nn.module.Layer import Layer
from analogvnn.utils.common_types import TENSOR_OPERABLE
from analogvnn.utils.to_tensor_parameter import to_float_tensor, to_nongrad_parameter

__all__ = ['LaplacianNoise']


class LaplacianNoise(Layer, BackwardIdentity):
    """Implements the Laplacian noise function.

    Attributes:
        scale (nn.Parameter): the scale of the Laplacian noise.
        leakage (nn.Parameter): the leakage of the Laplacian noise.
        precision (nn.Parameter): the precision of the Laplacian noise.
    """
    __constants__ = ['scale', 'leakage', 'precision']
    scale: nn.Parameter
    leakage: nn.Parameter
    precision: nn.Parameter

    def __init__(
            self,
            scale: Optional[float] = None,
            leakage: Optional[float] = None,
            precision: Optional[int] = None
    ):
        """initialize the Laplacian noise function.

        Args:
            scale (float): the scale of the Laplacian noise.
            leakage (float): the leakage of the Laplacian noise.
            precision (int): the precision of the Laplacian noise.
        """
        super(LaplacianNoise, self).__init__()

        if (scale is None) + (leakage is None) + (precision is None) != 1:
            raise ValueError("only 2 out of 3 arguments are needed (scale, leakage, precision)")

        scale, leakage, precision = to_float_tensor(scale, leakage, precision)

        if scale is None and leakage is not None and precision is not None:
            scale = self.calc_scale(leakage, precision)

        if precision is None and scale is not None and leakage is not None:
            precision = self.calc_precision(scale, leakage)

        if leakage is None and scale is not None and precision is not None:
            leakage = self.calc_leakage(scale, precision)

        self.scale, self.leakage, self.precision = to_nongrad_parameter(scale, leakage, precision)

    @staticmethod
    def calc_scale(leakage: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the scale of the Laplacian noise.

        Args:
            leakage (float): the leakage of the Laplacian noise.
            precision (int): the precision of the Laplacian noise.

        Returns:
            float: the scale of the Laplacian noise.
        """
        return - 1 / (2 * math.log(leakage) * precision)

    @staticmethod
    def calc_precision(scale: TENSOR_OPERABLE, leakage: TENSOR_OPERABLE) -> TENSOR_OPERABLE:
        """calculate the precision of the Laplacian noise.

        Args:
            scale (float): the scale of the Laplacian noise.
            leakage (float): the leakage of the Laplacian noise.

        Returns:
            int: the precision of the Laplacian noise.
        """
        return - 1 / (2 * math.log(leakage) * scale)

    @staticmethod
    def calc_leakage(scale: TENSOR_OPERABLE, precision: TENSOR_OPERABLE) -> Tensor:
        """calculate the leakage of the Laplacian noise.

        Args:
            scale (float): the scale of the Laplacian noise.
            precision (int): the precision of the Laplacian noise.

        Returns:
            float: the leakage of the Laplacian noise.
        """
        return 2 * LaplacianNoise.static_cdf(x=-1 / (2 * precision), scale=scale)

    @property
    def stddev(self) -> Tensor:
        """the standard deviation of the Laplacian noise.

        Returns:
            Tensor: the standard deviation of the Laplacian noise.
        """
        return (2 ** 0.5) * self.scale

    @property
    def variance(self) -> Tensor:
        """the variance of the Laplacian noise.

        Returns:
            Tensor: the variance of the Laplacian noise.
        """
        return 2 * self.scale.pow(2)

    def pdf(self, x: TENSOR_OPERABLE, loc: TENSOR_OPERABLE = 0) -> Tensor:
        """the probability density function of the Laplacian noise.

        Args:
            x (TENSOR_OPERABLE): the input tensor.
            loc (TENSOR_OPERABLE): the mean of the Laplacian noise.

        Returns:
            Tensor: the probability density function of the Laplacian noise.
        """
        return torch.exp(self.log_prob(x=x, loc=loc))

    def log_prob(self, x: TENSOR_OPERABLE, loc: TENSOR_OPERABLE = 0) -> Tensor:
        """the log probability density function of the Laplacian noise.

        Args:
            x (TENSOR_OPERABLE): the input tensor.
            loc (TENSOR_OPERABLE): the mean of the Laplacian noise.

        Returns:
            Tensor: the log probability density function of the Laplacian noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        loc = loc if isinstance(loc, Tensor) else torch.tensor(loc, requires_grad=False)
        return -torch.log(2 * self.scale) - torch.abs(x - loc) / self.scale

    @staticmethod
    def static_cdf(x: TENSOR_OPERABLE, scale: TENSOR_OPERABLE, loc: TENSOR_OPERABLE = 0.) -> TENSOR_OPERABLE:
        """the cumulative distribution function of the Laplacian noise.

        Args:
            x (TENSOR_OPERABLE): the input tensor.
            scale (TENSOR_OPERABLE): the scale of the Laplacian noise.
            loc (TENSOR_OPERABLE): the mean of the Laplacian noise.

        Returns:
            TENSOR_OPERABLE: the cumulative distribution function of the Laplacian noise.
        """
        return 0.5 - 0.5 * np.sign(x - loc) * np.expm1(-abs(x - loc) / scale)

    def cdf(self, x: Tensor, loc: Tensor = 0) -> Tensor:
        """the cumulative distribution function of the Laplacian noise.

        Args:
            x (Tensor): the input tensor.
            loc (Tensor): the mean of the Laplacian noise.

        Returns:
            Tensor: the cumulative distribution function of the Laplacian noise.
        """
        x = x if isinstance(x, Tensor) else torch.tensor(x, requires_grad=False)
        loc = loc if isinstance(loc, Tensor) else torch.tensor(loc, requires_grad=False)
        return self.static_cdf(x=x, scale=self.scale, loc=loc)

    def forward(self, x: Tensor) -> Tensor:
        """add Laplacian noise to the input tensor.

        Args:
            x (Tensor): the input tensor.

        Returns:
            Tensor: the output tensor with Laplacian noise.
        """
        return torch.distributions.Laplace(loc=x, scale=self.scale).sample()

    def extra_repr(self) -> str:
        """the extra representation of the Laplacian noise.

        Returns:
            str: the extra representation of the Laplacian noise.
        """
        return f'std={float(self.std):.4f}, leakage={float(self.leakage):.4f}, precision={int(self.precision)}'



from __future__ import annotations

import abc
from typing import Callable, Any, Optional, Sequence, Tuple

import torch
from torch import nn, Tensor, autograd

from analogvnn.utils.common_types import TENSORS

__all__ = ['BackwardModule']


class BackwardModule(abc.ABC):
    """Base class for all backward modules.

    A backward module is a module that can be used to compute the backward gradient of a given
    function. It is used to compute the gradient of the input of a function with respect to the output
    of the function.

    Attributes:
        _layer(Optional[nn.Module]): The layer for which the backward gradient is computed.
        _empty_holder_tensor(Tensor): A placeholder tensor which always requires gradient for backward gradient
         computation.
    """
    _layer: Optional[nn.Module]
    _empty_holder_tensor: Tensor

    # noinspection PyAbstractClass
    class AutogradBackward(autograd.Function):
        """Autograd function is used as an optimization and proper calculation of gradients
        when using the autograd engine.
        """

        # noinspection PyMethodOverriding
        @staticmethod
        def forward(ctx: Any, backward_module: BackwardModule, _: Tensor, *args: Tensor, **kwargs: Tensor) -> TENSORS:
            """Forward pass of the autograd function.

            Args:
                ctx: The context of the autograd function.
                backward_module(BackwardModule): The backward module.
                _ (Tensor): placeholder tensor which always requires grad.
                *args (Tensor): The arguments of the function.
                **kwargs (Tensor): The keyword arguments of the function.

            Returns:
                TENSORS: The output of the function.
            """
            ctx.backward_module = backward_module
            return ctx.backward_module._call_impl_forward(*args, **kwargs)

        @staticmethod
        def backward(ctx: Any, *grad_outputs: Tensor) -> Tuple[None, None, TENSORS]:
            """Backward pass of the autograd function.

            Args:
                ctx: The context of the autograd function.
                *grad_outputs (Tensor): The gradients of the output of the function.

            Returns:
                TENSORS: The gradients of the input of the function.
            """

            backward_module: BackwardModule = ctx.backward_module
            results = backward_module._call_impl_backward(*grad_outputs)

            if isinstance(results, Sequence):
                return (None, None, *results)

            return None, None, results

    def __init__(self, layer: nn.Module = None):
        """Initializes of the BackwardModule class.

        Args:
            layer (nn.Module): The layer for which the backward gradient is computed.
        """
        super(BackwardModule, self).__init__()
        self._layer = None
        self._empty_holder_tensor = torch.ones((1,), requires_grad=True)
        if not isinstance(self, nn.Module):
            self.set_layer(layer)

    def forward(self, *inputs: Tensor, **inputs_kwarg: Tensor) -> TENSORS:
        """Forward pass of the layer.

        Args:
            *inputs (Tensor): The inputs of the layer.
            **inputs_kwarg (Tensor): The keyword inputs of the layer.

        Returns:
            TENSORS: The output of the layer.

        Raises:
            NotImplementedError: If the forward pass is not implemented.
        """
        raise NotImplementedError(f"Module [{type(self).__name__}] is missing the required \"forward\" function")

    forward._implemented = False

    @abc.abstractmethod
    def backward(self, *grad_outputs: Tensor, **grad_output_kwarg: Tensor) -> TENSORS:
        """Backward pass of the layer.

        Args:
            *grad_outputs (Tensor): The gradients of the output of the layer.
            **grad_output_kwarg (Tensor): The keyword gradients of the output of the layer.

        Returns:
            TENSORS: The gradients of the input of the layer.

        Raises:
            NotImplementedError: If the backward pass is not implemented.
        """
        raise NotImplementedError(f"Module [{type(self).__name__}] is missing the required \"backward\" function")

    def _call_impl_forward(self, *args: Tensor, **kwarg: Tensor) -> TENSORS:
        """Calls Forward pass of the layer.

        Args:
            *inputs (Tensor): The inputs of the layer.
            **inputs_kwarg (Tensor): The keyword inputs of the layer.

        Returns:
            TENSORS: The output of the layer.
        """
        return self.forward(*args, **kwarg)

    def _call_impl_backward(self, *grad_output: Tensor, **grad_output_kwarg: Tensor) -> TENSORS:
        """Calls Backward pass of the layer.

        Args:
            *grad_outputs (Tensor): The gradients of the output of the layer.
            **grad_output_kwarg (Tensor): The keyword gradients of the output of the layer.

        Returns:
            TENSORS: The gradients of the input of the layer.
        """
        return self.backward(*grad_output, **grad_output_kwarg)

    __call__: Callable[..., Any] = _call_impl_backward

    def auto_apply(self, *args: Tensor, to_apply=True, **kwargs: Tensor) -> TENSORS:
        """Applies the backward module to the given layer using the proper method.

        Args:
            *args (Tensor): The inputs of the layer.
            to_apply (bool): if True and is training then the AutogradBackward is applied,
            otherwise the backward module is applied.
            **kwargs (Tensor): The keyword inputs of the layer.

        Returns:
            TENSORS: The output of the layer.
        """
        if to_apply and self.layer.training:
            return BackwardModule.AutogradBackward.apply(self, self._empty_holder_tensor, *args, **kwargs)
        else:
            return self._call_impl_forward(*args, **kwargs)

    def has_forward(self) -> bool:
        """Checks if the forward pass is implemented.

        Returns:
            bool: True if the forward pass is implemented, False otherwise.
        """
        return not hasattr(self.forward, "_implemented")

    @property
    def layer(self) -> nn.Module:
        """Gets the layer for which the backward gradient is computed.

        Returns:
            nn.Module: layer
        """
        return self.get_layer()

    def get_layer(self) -> nn.Module:
        """Gets the layer for which the backward gradient is computed.

        Returns:
            nn.Module: layer
        """
        if isinstance(self, nn.Module):
            return self
        else:
            return self._layer

    def set_layer(self, layer: Optional[nn.Module]) -> BackwardModule:
        """Sets the layer for which the backward gradient is computed.

        Args:
            layer (nn.Module): The layer for which the backward gradient is computed.

        Returns:
            BackwardModule: self

        Raises:
            ValueError: If self is a subclass of nn.Module.
            ValueError: If the layer is already set.
            ValueError: If the layer is not an instance of nn.Module.
        """
        if isinstance(self, nn.Module):
            raise ValueError(f"layer of Backward Module is set to itself")
        if self._layer is not None:
            raise ValueError(f"changing the layer of Backward Module is not allowed")
        if layer is not None and not isinstance(layer, nn.Module):
            raise ValueError(f'layer not instance of Layer class')

        self._layer = layer
        return self

    @staticmethod
    def set_grad_of(tensor: Tensor, grad: Tensor) -> Optional[Tensor]:
        """Sets the gradient of the given tensor.

        Args:
            tensor (Tensor): The tensor.
            grad (Tensor): The gradient.

        Returns:
            Optional[Tensor]: the gradient of the tensor.
        """
        if tensor is None or tensor.requires_grad is False:
            return None

        # noinspection PyBroadException
        try:
            tensor.backward(gradient=grad, inputs=tensor)
        except Exception:
            # noinspection PyProtectedMember
            for key, value in tensor._backward_hooks.items():
                grad = value(grad)

            if tensor.grad is None:
                tensor.grad = grad
            else:
                tensor.grad += grad

        return tensor.grad

    def __getattr__(self, name: str) -> Any:
        """Gets the attribute of the layer.

        Args:
            name (str): The name of the attribute.

        Returns:
            Any: The attribute of the layer.

        Raises:
            AttributeError: If the attribute is not found.
        """
        if isinstance(self, nn.Module) or self == self._layer:
            return super(BackwardModule, self).__getattr__(name)
        if not str(name).startswith("__") and self._layer is not None and hasattr(self._layer, name):
            return getattr(self._layer, name)
        raise AttributeError("'{}' object has no attribute '{}'".format(type(self).__name__, name))



from abc import ABC

from torch import Tensor

from analogvnn.backward.BackwardModule import BackwardModule
from analogvnn.utils.common_types import TENSORS

__all__ = ['BackwardUsingForward']


class BackwardUsingForward(BackwardModule, ABC):
    """The backward module that uses the forward function to compute the backward gradient.
    """

    def backward(self, *grad_output: Tensor, **grad_output_kwarg: Tensor) -> TENSORS:
        """Computes the backward gradient of the input of the layer with respect to the output of the layer
        using the forward function.

        Args:
            *grad_output (Tensor): The gradients of the output of the layer.
            **grad_output_kwarg (Tensor): The gradients of the output of the layer.

        Returns:
            TENSORS: The gradients of the input of the layer.
        """
        return self._layer.forward(*grad_output, **grad_output_kwarg)
